use std::fs::{File, OpenOptions};

use bevy::prelude::*;
use bevy::reflect::{
    ArrayInfo, EnumInfo, ListInfo, MapInfo, NamedField, OpaqueInfo, SetInfo, StructInfo,
    StructVariantInfo, TupleInfo, TupleStructInfo, TupleVariantInfo, Type, TypeInfo,
    TypeRegistration, UnitVariantInfo, UnnamedField, VariantInfo,
};
use bevy::utils::HashMap;
use serde::{Deserialize, Serialize};

use crate::BevyBlenderConfig;

pub(crate) struct CreateRegistryPlugin;

impl Plugin for CreateRegistryPlugin {
    fn build(&self, app: &mut bevy::prelude::App) {
        app.add_systems(Startup, create_registry);
    }
}

fn create_registry(config: Res<BevyBlenderConfig>, type_registry: Res<AppTypeRegistry>) {
    let type_registry = type_registry.read();
    let filter = &config.type_filter;
    let types: Vec<RegistryEntry> = type_registry
        .iter()
        // TODO: somehow filter by components?
        //.filter(|t| is_component(t))
        .filter(|t| filter.is_allowed_by_id(t.type_id()))
        .map(RegistryEntry::from)
        .collect();

    if let Ok(file) = File::create(&config.registry_path) {
        if let Err(e) = serde_json::to_writer(&file, &types) {
            warn!(
                "Failed to write registry to {:?}: {e:?}",
                &config.registry_path
            );
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
enum SchemaType {
    Array,
    Object,
    Value(ValueVariant),
}

impl From<&NativeType> for SchemaType {
    fn from(value: &NativeType) -> Self {
        match value {
            NativeType::Enum { .. } => SchemaType::Object,
            NativeType::List { .. } => SchemaType::Array,
            NativeType::Map { .. } => SchemaType::Object,
            NativeType::Opaque(value) => SchemaType::Value(ValueVariant::from(value)),
            NativeType::Set { .. } => SchemaType::Object,
            NativeType::Tuple { .. } => SchemaType::Array,
            NativeType::Struct { .. } => SchemaType::Object,
            NativeType::Unit => SchemaType::Object,
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
enum ValueVariant {
    Boolean,
    UnsignedInteger,
    Integer,
    Float,
    String,
    Object,
}

impl From<&OpaqueVariant> for ValueVariant {
    fn from(value: &OpaqueVariant) -> Self {
        match value {
            OpaqueVariant::Boolean => Self::Boolean,
            OpaqueVariant::Char => Self::String,
            OpaqueVariant::Integer(IntSign::Unsigned, _) => Self::UnsignedInteger,
            OpaqueVariant::Integer(IntSign::Signed, _) => Self::Integer,
            OpaqueVariant::Float => Self::Float,
            OpaqueVariant::String => Self::String,
            OpaqueVariant::Struct => Self::Object,
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct TypeRef {
    #[serde(rename = "$ref")]
    ref_type: String,
}

impl TypeRef {
    pub fn new<T: std::fmt::Display>(value: T) -> Self {
        Self {
            ref_type: format!("#/defs/{value}"),
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct FieldType {
    #[serde(rename = "type")]
    type_ref: TypeRef,
}

impl FieldType {
    pub fn new<T: std::fmt::Display>(value: T) -> Self {
        Self {
            type_ref: TypeRef::new(value),
        }
    }
}

impl From<Type> for FieldType {
    fn from(value: Type) -> Self {
        Self::new(value.path())
    }
}

impl From<&Type> for FieldType {
    fn from(value: &Type) -> Self {
        Self::new(value.path())
    }
}

impl From<&NamedField> for FieldType {
    fn from(value: &NamedField) -> Self {
        value.ty().into()
    }
}

impl From<&UnnamedField> for FieldType {
    fn from(value: &UnnamedField) -> Self {
        value.ty().into()
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
enum NativeType {
    Enum {
        one_of: Vec<NativeType>,
    },
    List {
        items: FieldType,
        variant: ListVariant,
    },
    Map {
        key_type: FieldType,
        value_type: FieldType,
    },
    Opaque(OpaqueVariant),
    Set {
        value_type: FieldType,
    },
    Tuple {
        prefix_items: Vec<FieldType>,
        variant: TupleVariant,
    },
    Struct {
        properties: HashMap<String, FieldType>,
        additional_properties: bool,
        required: Vec<String>,
    },
    Unit,
}

enum StructType<'a> {
    Struct(&'a StructInfo),
    StructVariant(&'a StructVariantInfo),
}

enum TupleType<'a> {
    Tuple(&'a TupleInfo),
    TupleVariant(&'a TupleVariantInfo),
}

impl NativeType {
    fn from_struct(value: StructType) -> Self {
        let iter = match value {
            StructType::Struct(info) => info.iter(),
            StructType::StructVariant(info) => info.iter(),
        };

        let properties = iter
            .as_ref()
            .into_iter()
            .map(|field| (field.name().to_string(), field.into()))
            .collect();

        let option_path = "core::option::Option";
        let required = iter
            .filter(|field| !field.type_path().starts_with(&option_path))
            .map(|field| field.name().to_string())
            .collect();

        Self::Struct {
            properties,
            additional_properties: false,
            required,
        }
    }

    fn from_tuple(value: TupleType) -> Self {
        let iter = match value {
            TupleType::Tuple(info) => info.iter(),
            TupleType::TupleVariant(info) => info.iter(),
        };

        let prefix_items = iter.map(FieldType::from).collect();

        Self::Tuple {
            prefix_items,
            variant: TupleVariant::Tuple,
        }
    }
}

impl From<&VariantInfo> for NativeType {
    fn from(value: &VariantInfo) -> Self {
        match value {
            VariantInfo::Struct(info) => NativeType::from(info),
            VariantInfo::Tuple(info) => NativeType::from(info),
            VariantInfo::Unit(info) => NativeType::from(info),
        }
    }
}

impl From<&StructVariantInfo> for NativeType {
    fn from(value: &StructVariantInfo) -> Self {
        Self::from_struct(StructType::StructVariant(value))
    }
}

impl From<&TupleVariantInfo> for NativeType {
    fn from(value: &TupleVariantInfo) -> Self {
        Self::from_tuple(TupleType::TupleVariant(value))
    }
}

impl From<&UnitVariantInfo> for NativeType {
    fn from(_value: &UnitVariantInfo) -> Self {
        Self::Unit
    }
}

impl From<&TypeRegistration> for NativeType {
    fn from(value: &TypeRegistration) -> Self {
        value.type_info().into()
    }
}

impl From<&TypeInfo> for NativeType {
    fn from(value: &TypeInfo) -> Self {
        match value {
            TypeInfo::Struct(info) => NativeType::from(info),
            TypeInfo::TupleStruct(info) => NativeType::from(info),
            TypeInfo::Tuple(info) => NativeType::from(info),
            TypeInfo::List(info) => NativeType::from(info),
            TypeInfo::Array(info) => NativeType::from(info),
            TypeInfo::Map(info) => NativeType::from(info),
            TypeInfo::Set(info) => NativeType::from(info),
            TypeInfo::Enum(info) => NativeType::from(info),
            TypeInfo::Opaque(info) => NativeType::from(info),
        }
    }
}

impl From<&ArrayInfo> for NativeType {
    fn from(value: &ArrayInfo) -> Self {
        Self::List {
            items: value.item_ty().into(),
            variant: ListVariant::Array,
        }
    }
}

impl From<&EnumInfo> for NativeType {
    fn from(value: &EnumInfo) -> Self {
        let one_of = value.iter().map(NativeType::from).collect();

        Self::Enum { one_of }
    }
}

impl From<&ListInfo> for NativeType {
    fn from(value: &ListInfo) -> Self {
        Self::List {
            items: FieldType::new(value.item_ty().path()),
            variant: ListVariant::List,
        }
    }
}

impl From<&MapInfo> for NativeType {
    fn from(value: &MapInfo) -> Self {
        Self::Map {
            key_type: FieldType::new(value.key_ty().path()),
            value_type: FieldType::new(value.value_ty().path()),
        }
    }
}

impl From<&OpaqueInfo> for NativeType {
    fn from(value: &OpaqueInfo) -> Self {
        Self::Opaque(OpaqueVariant::from(value))
    }
}

impl From<&SetInfo> for NativeType {
    fn from(value: &SetInfo) -> Self {
        Self::Set {
            value_type: value.value_ty().into(),
        }
    }
}

impl From<&StructInfo> for NativeType {
    fn from(value: &StructInfo) -> Self {
        Self::from_struct(StructType::Struct(value))
    }
}

impl From<&TupleInfo> for NativeType {
    fn from(value: &TupleInfo) -> Self {
        Self::from_tuple(TupleType::Tuple(value))
    }
}

impl From<&TupleStructInfo> for NativeType {
    fn from(value: &TupleStructInfo) -> Self {
        let prefix_items = value.iter().map(FieldType::from).collect();
        Self::Tuple {
            prefix_items,
            variant: TupleVariant::TupleStruct,
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
enum ListVariant {
    Array,
    List,
}

#[derive(Debug, Serialize, Deserialize)]
enum TupleVariant {
    Tuple,
    TupleStruct,
}

#[derive(Debug, Serialize, Deserialize)]
enum IntSign {
    Signed,
    Unsigned,
}

impl IntSign {
    fn from_str<T: AsRef<str>>(value: T) -> Self {
        if value.as_ref().starts_with('u') {
            IntSign::Unsigned
        } else {
            IntSign::Signed
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
enum IntSize {
    Sized(u8),
    PointerSized,
}

impl IntSize {
    fn from_str<T: AsRef<str>>(value: T) -> Self {
        let size = &value.as_ref()[1..];
        if size == "size" {
            Self::PointerSized
        } else {
            let size = u8::from_str_radix(size, 10).unwrap();
            Self::Sized(size)
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
enum OpaqueVariant {
    Boolean,
    Integer(IntSign, IntSize),
    Float,
    Char,
    String,
    Struct,
}

impl OpaqueVariant {
    fn int_from_str<T: AsRef<str>>(value: T) -> Self {
        let sign = IntSign::from_str(&value);
        let size = IntSize::from_str(&value);

        Self::Integer(sign, size)
    }
}

impl From<&OpaqueInfo> for OpaqueVariant {
    fn from(value: &OpaqueInfo) -> Self {
        match value.type_path() {
            "bool" => Self::Boolean,
            value @ ("u8" | "u16" | "u32" | "u64" | "u128" | "usize" | "i8" | "i16" | "i32"
            | "i64" | "i128" | "isize") => Self::int_from_str(value),
            "f32" | "f64" => Self::Float,
            "char" | "str" | "alloc::string::String" => Self::String,
            _ => Self::Struct,
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct RegistryEntry {
    #[serde(rename = "type")]
    schema_type: SchemaType,
    native_type: NativeType,
    qualified_type_path: String,
}

impl From<&TypeRegistration> for RegistryEntry {
    fn from(value: &TypeRegistration) -> Self {
        let type_info = value.type_info();
        let native_type = NativeType::from(type_info);
        let schema_type = SchemaType::from(&native_type);

        Self {
            schema_type,
            native_type,
            qualified_type_path: type_info.type_path().to_string(),
        }
    }
}
